import io
import zipfile
from os import makedirs, rmdir
from os.path import isdir, join
from shutil import move
from glob import iglob

from minet.web import request
from minet.exceptions import InvalidStatusError
from minet.browser.utils import get_browsers_path
from minet.loggers import downloaders_logger

AVAILABLE_EXTENSIONS = {
    "i-still-dont-care-about-cookies": {
        "url": "https://github.com/OhMyGuus/I-Still-Dont-Care-About-Cookies/releases/download/v1.1.1/ISDCAC-chrome-source.zip",
        "root": "./",
    },
    "ublock-origin": {
        "url": "https://github.com/gorhill/uBlock/releases/download/1.52.2/uBlock0_1.52.2.chromium.zip",
        "root": "./uBlock0.chromium",
    },
}


def ensure_extension_is_downloaded(name: str) -> bool:
    if name not in AVAILABLE_EXTENSIONS:
        raise TypeError("unknown extension %s" % name)

    extension_info = AVAILABLE_EXTENSIONS[name]

    browsers_dir = get_browsers_path()
    extensions_dir = join(browsers_dir, "extensions")
    extension_dir = join(extensions_dir, name)

    makedirs(extensions_dir, exist_ok=True)

    if isdir(extension_dir):
        return True

    downloaders_logger.info(
        "Downloading browser extension: %s" % name,
        extra={"namespace": "extension", "target": name},
    )
    response = request(extension_info["url"])

    if response.status != 200:
        raise InvalidStatusError(response.status)

    with zipfile.ZipFile(io.BytesIO(response.body)) as z:
        z.extractall(extension_dir)

    root = extension_info["root"]

    if root != "./":
        for p in iglob(join(extension_dir, root, "*")):
            move(p, extension_dir)

        rmdir(join(extension_dir, root))

    return False


def get_extension_path(name: str) -> str:
    if name not in AVAILABLE_EXTENSIONS:
        raise TypeError("unknown extension %s" % name)

    return join(get_browsers_path(), "extensions", name)


if __name__ == "__main__":
    print("Downloading missing extensions...\n")

    for name in AVAILABLE_EXTENSIONS:
        print("  -> %s" % name)
        already_downloaded = ensure_extension_is_downloaded(name)
        print("     Already ok." if already_downloaded else "     Acquired.")
        print()



from typing import Iterable, List, Set, Iterator, Any, Optional, Literal

from dataclasses import dataclass
from ural.lru import CanonicalizedLRUTrie
from ural import links_from_html, could_be_html, infer_redirection, canonicalize_url
from casanova import TabularRecord
from functools import lru_cache

from minet.crawl.spiders import Spider, SpiderResult
from minet.crawl.types import CrawlJob, SuccessfulCrawlResult
from minet.web import Response, resolve

VALID_WEBENTITY_STATUSES = ["IN", "OUT", "UNDECIDED", "DISCOVERED"]

WebentityStatus = Literal["IN", "OUT", "UNDECIDED", "DISCOVERED"]


@dataclass
class WebentityRecord:
    id: str
    status: WebentityStatus


@dataclass
class WebentityLink(TabularRecord):
    source_webentity: str
    target_webentity: str
    source_url: str
    target_url: str


@dataclass
class HypheSpiderAddendum:
    webentity_id: str
    links: List[WebentityLink]


def extract_links(url: str, body: bytes, encoding: str) -> List[str]:
    return list(
        links_from_html(
            url,
            body,
            encoding=encoding,
            canonicalize=True,
            unique=True,
            strip_fragment=True,
        )
    )


@lru_cache(2048)
def resolve_shortened_link(url: str) -> str:
    try:
        stack = resolve(url)
        last_url = stack[-1].url
        return last_url
    except Exception:
        return url


class HypheSpider(Spider):
    def __init__(self, ignore_internal_links: bool = False):
        self.trie: CanonicalizedLRUTrie[WebentityRecord] = CanonicalizedLRUTrie(
            suffix_aware=True, strip_fragment=True
        )
        self.start_pages: Set[str] = set()
        self.ignore_internal_links = ignore_internal_links

    def start(self) -> Iterable[str]:
        yield from self.start_pages

    def add_start_page(self, url: str) -> None:
        if self.is_attached:
            raise RuntimeError("cannot add start page if spider is already attached")

        self.start_pages.add(canonicalize_url(url, strip_fragment=True))

    def set(self, prefix, webentity, status: WebentityStatus = "IN") -> None:
        if status not in VALID_WEBENTITY_STATUSES:
            raise TypeError("invalid webentity status: {!r}".format(status))

        self.trie.set(prefix, WebentityRecord(id=webentity, status=status))

    def process(self, job: CrawlJob, response: Response) -> SpiderResult:
        if response.status != 200:
            return

        if not response.is_html:
            return

        webentity = self.trie.match(response.end_url)

        if webentity is None or webentity.status != "IN":
            return

        urls = self.submit(
            extract_links, response.end_url, response.body, response.likely_encoding
        )

        urls_to_follow = []
        links: List[WebentityLink] = []

        for url in urls:
            url = infer_redirection(url)

            if not could_be_html(url):
                continue

            # if is_shortened_url(url):
            #     url = resolve_shortened_link(url)

            match = self.trie.match(url)

            if match is None:
                continue

            if match.status != "IN":
                continue

            if not self.ignore_internal_links or webentity.id != match.id:
                links.append(
                    WebentityLink(
                        source_webentity=webentity.id,
                        target_webentity=match.id,
                        source_url=response.end_url,
                        target_url=url,
                    )
                )

            if match.id != webentity.id:
                continue

            urls_to_follow.append(url)

        return (
            HypheSpiderAddendum(webentity_id=webentity.id, links=links),
            urls_to_follow,
        )

    def tabulate(
        self, result: SuccessfulCrawlResult[Any, Optional[HypheSpiderAddendum]]
    ) -> Iterator[WebentityLink]:
        if not result.data:
            return

        yield from result.data.links



# =============================================================================
# Minet Headers Utils
# =============================================================================
#
# Miscellaneous functions related to HTTP headers.
#
from typing import Optional, Tuple

import cgi


def parse_http_header(header: str) -> Tuple[str, str]:
    key, value = header.split(":", 1)

    return key.strip(), value.strip()


def get_encoding_from_content_type_header(
    content_type_header: Optional[str],
) -> Optional[str]:
    if content_type_header is None:
        return None

    parsed_header = cgi.parse_header(content_type_header)

    if len(parsed_header) > 1:
        charset = parsed_header[1].get("charset")

        if charset is not None:
            return charset.lower()

    return None


# TODO: take more cases into account...
#   http://www.otsukare.info/2015/03/26/refresh-http-header
def parse_http_refresh(value) -> Optional[Tuple[int, str]]:
    try:
        if isinstance(value, bytes):
            value = value.decode()

        duration, url = value.strip().split(";", 1)

        if not url.lower().strip().startswith("url="):
            return None

        return int(duration), str(url.split("=", 1)[1])
    except Exception:
        return None



from urllib.parse import urljoin
from ural import could_be_rss
from bs4 import SoupStrainer, BeautifulSoup

from .types import NamedScraper


class RssScraper(NamedScraper):
    name = "rss"
    fieldnames = ["rss_url"]
    plural = True
    output_type = "list"
    strainer = SoupStrainer(name=["a", "link"])

    def scrape(self, soup: BeautifulSoup, context=None):
        rss_urls = []
        base_url = context.get("url") if context is not None else ""

        for link in soup.find_all():
            if link.name == "link":
                type_attr = link.attrs.get("type", None)
                if (
                    type_attr == "application/rss+xml"
                    or type_attr == "application/atom+xml"
                ):
                    href = link.attrs.get("href", None)
                    if href:
                        rss_urls.append(urljoin(base_url, href))
            else:
                href = link.attrs.get("href", None)
                url = urljoin(base_url, href)
                if could_be_rss(url):
                    rss_urls.append(url)

        return rss_urls



# =============================================================================
# Minet Twitter Exceptions
# =============================================================================
#
from minet.exceptions import MinetError


class TwitterError(MinetError):
    pass


class TwitterPublicAPIError(TwitterError):
    pass


class TwitterPublicAPIGuestTokenError(TwitterPublicAPIError):
    pass


class TwitterPublicAPIInvalidCookieError(TwitterPublicAPIError):
    pass


class TwitterPublicAPIQueryTooLongError(TwitterPublicAPIError):
    pass


class TwitterPublicAPIRateLimitError(TwitterPublicAPIError):
    pass


class TwitterPublicAPIBadAuthError(TwitterPublicAPIError):
    def __init__(self, status: int):
        self.status = status
        super().__init__(str(status))


class TwitterPublicAPIInvalidResponseError(TwitterPublicAPIError):
    def __init__(self, message=None, response_text=None, status=None):
        super().__init__(message)
        self.response_text = response_text
        self.status = status

    def format_epilog(self):
        epilog = []

        if self.status is not None:
            epilog.append("Status: %i" % self.status)

        if self.response_text is not None:
            epilog.append("Response: %s" % self.response_text)

        return ", ".join(epilog) if epilog else ""


class TwitterPublicAPIncompleteTweetIndexError(TwitterPublicAPIError):
    def __init__(self, message=None, tweet_id=None, tweet_index=None):
        super().__init__(message)
        self.tweet_id = tweet_id
        self.tweet_index = tweet_index

    def format_epilog(self):
        epilog = []

        if self.tweet_id is not None:
            epilog.append("Tweet id: %s" % self.tweet_id)

        if self.tweet_index is not None:
            epilog.append("Tweets indexed: %s" % list(self.tweet_index.keys()))

        return ", ".join(epilog) if epilog else ""


class TwitterPublicAPIncompleteUserIndexError(TwitterPublicAPIError):
    def __init__(self, message=None, user_id=None, user_index=None):
        super().__init__(message)
        self.user_id = user_id
        self.user_index = user_index

    def format_epilog(self):
        epilog = []

        if self.user_id is not None:
            epilog.append("User id: %s" % self.user_id)

        if self.user_index is not None:
            epilog.append("Users indexed: %s" % list(self.user_index.keys()))

        return ", ".join(epilog) if epilog else ""


class TwitterPublicAPIBadRequest(TwitterPublicAPIError):
    pass


class TwitterPublicAPIOverCapacityError(TwitterPublicAPIError):
    pass


class TwitterPublicAPIParsingError(TwitterPublicAPIError):
    pass


class TwitterPublicAPIHiccupError(TwitterPublicAPIError):
    pass


class TwitterPublicAPINotWorkingAnymore(TwitterPublicAPIError):
    pass


class TwitterPublicAPINotFound(TwitterPublicAPIError):
    pass



from typing import Optional, List

from dataclasses import dataclass
from casanova import TabularRecord
from ebbe import getpath

from minet.dates import timestamp_to_isoformat


@dataclass
class TiktokVideo(TabularRecord):
    id: str
    description: str
    timestamp: int
    utc_time: str
    original_item: bool
    video_id: str
    video_height: int
    video_width: int
    video_duration: int
    video_ratio: str
    video_cover: str
    video_origin_cover: str
    video_dynamic_cover: str
    video_reflow_cover: str
    video_bitrate: int
    video_format: str
    video_quality: str
    author_id: str
    author_unique_id: str
    author_nickname: str
    author_avatar: str
    author_signature: str
    author_verified: bool
    author_secret: bool
    author_following_count: int
    author_follower_count: int
    author_heart_count: int
    author_video_count: int
    author_digg_count: int
    music_id: str
    music_title: str
    music_author_name: str
    music_original: bool
    music_play_url: str
    music_duration: int
    music_album: Optional[str]
    digg_count: int
    share_count: int
    comment_count: int
    play_count: int
    duet_from_video_id: Optional[str]
    duet_from_user_id: Optional[str]
    duet_from_user_name: Optional[str]
    duet_from_user_is_commerce: Optional[bool]
    challenges_titles: List[str]
    challenges_descriptions: List[str]
    stickers_texts: List[str]
    effect_stickers_names: List[str]
    duet_enabled: bool
    stitch_enabled: bool
    share_enabled: bool
    is_ad: bool

    @classmethod
    def from_payload(cls, payload) -> "TiktokVideo":
        challenges_titles = []
        challenges_descs = []
        stickers_texts = []
        effect_stickers_names = []

        challenges = payload.get("challenges")
        if challenges:
            for challenge in challenges:
                challenges_titles.append(challenge.get("title"))
                challenges_descs.append(challenge.get("desc"))

        stickers = payload.get("stickersOnItem")
        if stickers:
            for sticker in stickers:
                stickers_texts.append(" ".join(sticker.get("stickerText")))

        effects = payload.get("effectStickers")
        if effects:
            for effect in effects:
                effect_stickers_names.append(effect.get("name"))

        duet_from_id = getpath(payload, ["duetInfo", "duetFromId"])
        duet_from_user_id = None
        duet_from_user_name = None
        duet_from_user_is_commerce = None

        if duet_from_id == "0":
            duet_from_id = None
            duet_from_user_id = None
            duet_from_user_name = None
            duet_from_user_is_commerce = None
        else:
            for text in getpath(payload, ["textExtra"]):
                if text["awemeId"] == duet_from_id:
                    duet_from_user_id = text["userId"]
                    duet_from_user_name = text["userUniqueId"]
                    duet_from_user_is_commerce = text["isCommerce"]

        return cls(
            payload.get("id"),
            payload.get("desc"),
            payload.get("createTime"),
            timestamp_to_isoformat(payload.get("createTime")),
            payload.get("originalItem"),
            getpath(payload, ["video", "id"]),
            getpath(payload, ["video", "height"]),
            getpath(payload, ["video", "width"]),
            getpath(payload, ["video", "duration"]),
            getpath(payload, ["video", "ratio"]),
            getpath(payload, ["video", "cover"]),
            getpath(payload, ["video", "originCover"]),
            getpath(payload, ["video", "dynamicCover"]),
            getpath(payload, ["video", "reflowCover"]),
            getpath(payload, ["video", "bitrate"]),
            getpath(payload, ["video", "format"]),
            getpath(payload, ["video", "videoQuality"]),
            getpath(payload, ["author", "id"]),
            getpath(payload, ["author", "uniqueId"]),
            getpath(payload, ["author", "nickname"]),
            getpath(payload, ["author", "avatarLarger"]),
            getpath(payload, ["author", "signature"]),
            getpath(payload, ["author", "verified"]),
            getpath(payload, ["author", "secret"]),
            getpath(payload, ["authorStats", "followingCount"]),
            getpath(payload, ["authorStats", "followerCount"]),
            getpath(payload, ["authorStats", "heartCount"]),
            getpath(payload, ["authorStats", "videoCount"]),
            getpath(payload, ["authorStats", "diggCount"]),
            getpath(payload, ["music", "id"]),
            getpath(payload, ["music", "title"]),
            getpath(payload, ["music", "authorName"]),
            getpath(payload, ["music", "original"]),
            getpath(payload, ["music", "playUrl"]),
            getpath(payload, ["music", "duration"]),
            getpath(payload, ["music", "album"]),
            getpath(payload, ["stats", "diggCount"]),
            getpath(payload, ["stats", "shareCount"]),
            getpath(payload, ["stats", "commentCount"]),
            getpath(payload, ["stats", "playCount"]),
            duet_from_id,
            duet_from_user_id,
            duet_from_user_name,
            duet_from_user_is_commerce,
            challenges_titles,
            challenges_descs,
            stickers_texts,
            effect_stickers_names,
            payload.get("duetEnabled"),
            payload.get("stitchEnabled"),
            payload.get("shareEnabled"),
            payload.get("isAd"),
        )



from typing import (
    Callable,
    TypeVar,
    Awaitable,
    Set,
    Optional,
    Union,
    Literal,
    Container,
)
from minet.types import Concatenate, ParamSpec

import os
import asyncio
from shutil import rmtree
from concurrent.futures import Future
from threading import Thread, Event, Lock
from playwright.async_api import (
    async_playwright,
    Browser,
    BrowserContext,
    Error as PlaywrightError,
    TimeoutError as PlaywrightTimeoutError,
)
from urllib3._collections import HTTPHeaderDict

from minet.constants import REDIRECT_STATUSES
from minet.exceptions import (
    UnknownBrowserError,
    BrowserYetUnimplementedError,
    InvalidStatusError,
)
from minet.web import Response
from minet.browser.plawright_shim import install_browser
from minet.browser.utils import (
    get_browsers_path,
    get_temp_persistent_context_path,
    convert_playwright_error,
)
from minet.browser.extensions import get_extension_path, ensure_extension_is_downloaded
from minet.types import Redirection

SUPPORTED_BROWSERS = ("chromium", "firefox")


BrowserName = Literal["chromium", "firefox"]
T = TypeVar("T")
P = ParamSpec("P")
BrowserOrBrowserContext = Union[Browser, BrowserContext]
BrowserCallable = Callable[Concatenate[Browser, P], Awaitable[T]]
BrowserContextCallable = Callable[Concatenate[BrowserContext, P], Awaitable[T]]
BrowserOrBrowserContextCallable = Callable[
    Concatenate[BrowserOrBrowserContext, P], Awaitable[T]
]

# TODO: introduce stealth back
# TODO: abstract screenshot function
# TODO: helper for stealth page, monkey patch some function?


class ThreadsafeBrowser:
    def __init__(
        self,
        browser: BrowserName = "chromium",
        headless: bool = True,
        automatic_consent: bool = False,
        adblock: bool = False,
        width: int = 1920,
        height: int = 1080,
    ) -> None:
        if browser not in SUPPORTED_BROWSERS:
            raise UnknownBrowserError(browser)

        self.extensions = []

        if automatic_consent:
            self.extensions.append("i-still-dont-care-about-cookies")

        if adblock:
            self.extensions.append("ublock-origin")

        # NOTE: doing this in constructor to avoid threading issues down the line
        for ext in self.extensions:
            ensure_extension_is_downloaded(ext)

        self.requires_extensions = bool(self.extensions)
        self.persistent_user_data_dir = None
        self.persistent = self.requires_extensions

        if self.requires_extensions and browser != "chromium":
            raise TypeError("adblock and automatic_consent only work with chromium")

        self.browser_name: str = browser
        self.browser: Optional[Browser] = None
        self.default_browser_context: Optional[BrowserContext] = None
        self.headless = headless
        self.width = width
        self.height = height

        os.environ.setdefault("PLAYWRIGHT_BROWSERS_PATH", get_browsers_path())

        install_browser(self.browser_name)

        self.loop = asyncio.new_event_loop()
        self.start_event = Event()
        self.thread = Thread(
            name="Thread-browser-%i" % id(self), target=self.__thread_worker
        )

        self.running_futures: Set[Future] = set()
        self.running_futures_lock = Lock()

        # Starting loop thread
        self.thread.start()
        self.start_event.wait()

    async def __start_playwright(self) -> None:
        self.playwright = await async_playwright().start()

        if self.browser_name == "chromium":
            browser_type = self.playwright.chromium
        elif self.browser_name == "firefox":
            browser_type = self.playwright.firefox
        else:
            raise TypeError("unsupported browser")

        if self.requires_extensions:
            args = []
            disable_extensions_except = []

            for ext in self.extensions:
                p = get_extension_path(ext)
                disable_extensions_except.append(p)
                args.append(f"--load-extension={p}")

            args.append(
                f"--disable-extensions-except={','.join(disable_extensions_except)}"
            )

            if self.headless:
                args.append("--headless=new")

            user_data_dir = get_temp_persistent_context_path()
            rmtree(user_data_dir, ignore_errors=True)
            self.persistent_user_data_dir = user_data_dir

            self.default_browser_context = await browser_type.launch_persistent_context(
                user_data_dir,
                headless=False,
                args=args,
                viewport={"width": self.width, "height": self.height},
                locale="en-US",
            )

        else:
            self.browser = await browser_type.launch(headless=self.headless)
            self.default_browser_context = await self.browser.new_context(
                viewport={"width": self.width, "height": self.height}, locale="en-US"
            )

    async def __stop_playwright(self) -> None:
        # NOTE: we need to make sure those were actually launched, in
        # case of a nasty race condition

        if self.default_browser_context is not None:
            await self.default_browser_context.close()

        if self.browser is not None:
            await self.browser.close()

        # NOTE: this hangs without the proper child watcher
        await self.playwright.stop()

    def stop(self) -> None:
        # NOTE: if we don't do this and some job sent
        # to a threadpool executor raises and triggers
        # the closing of the playwright driver, then a
        # catastrophic chain reaction of exception will
        # make some jobs hang up and block indefinitely
        # which will cause a deadlock.
        with self.running_futures_lock:
            for fut in self.running_futures:
                if not fut.done():
                    fut.cancel()

        self.loop.call_soon_threadsafe(self.loop.stop)
        self.thread.join()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.stop()

    def __thread_worker(self):
        asyncio.set_event_loop(self.loop)
        self.loop.run_until_complete(self.__start_playwright())
        self.start_event.set()

        # NOTE: we are now ready to accept tasks
        try:
            self.loop.run_forever()
        finally:
            self.loop.run_until_complete(self.loop.shutdown_asyncgens())

        self.loop.run_until_complete(self.__stop_playwright())

    def __handle_future(self, future: Future):
        with self.running_futures_lock:
            self.running_futures.add(future)

        try:
            return future.result()
        finally:
            with self.running_futures_lock:
                self.running_futures.remove(future)

    def run(self, fn: BrowserCallable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
        if self.browser is None:
            raise TypeError("cannot run browser-level callable with persistent context")

        future = asyncio.run_coroutine_threadsafe(
            fn(self.browser, *args, **kwargs), self.loop
        )

        return self.__handle_future(future)

    def run_in_default_context(
        self, fn: BrowserContextCallable[P, T], *args: P.args, **kwargs: P.kwargs
    ) -> T:
        assert self.default_browser_context is not None

        future = asyncio.run_coroutine_threadsafe(
            fn(self.default_browser_context, *args, **kwargs), self.loop
        )

        return self.__handle_future(future)

    def run_in_browser_or_default_context(
        self,
        fn: BrowserOrBrowserContextCallable[P, T],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        target = self.browser

        if target is None:
            target = self.default_browser_context

        assert target is not None

        future = asyncio.run_coroutine_threadsafe(
            fn(target, *args, **kwargs), self.loop
        )

        return self.__handle_future(future)

    async def __request(
        self,
        context: BrowserContext,
        url: str,
        raise_on_statuses: Optional[Container[int]] = None,
    ) -> Response:
        async with await context.new_page() as page:
            try:
                emulated_response = await page.goto(url)
            except (PlaywrightError, PlaywrightTimeoutError) as e:
                error = convert_playwright_error(e)

                if isinstance(error, BrowserYetUnimplementedError):
                    raise e

                raise error

            assert emulated_response is not None

            # Invalid status?
            if (
                raise_on_statuses is not None
                and emulated_response.status in raise_on_statuses
            ):
                raise InvalidStatusError(emulated_response.status)

            # Collection redirections
            responses = []

            current = emulated_response

            while True:
                responses.append(current)

                if not current.request.redirected_from:
                    break

                current = await current.request.redirected_from.response()

                assert current is not None

            redirection_stack = []

            for r in reversed(responses):
                redirection_stack.append(
                    Redirection(
                        r.url,
                        "hit"
                        if r.status not in REDIRECT_STATUSES
                        else "location-header",
                        status=r.status,
                    )
                )

            # Building headers
            headers = HTTPHeaderDict()

            for header in await emulated_response.headers_array():
                headers[header["name"]] = header["value"]

            # Building response
            response = Response(
                url,
                redirection_stack,
                headers,
                emulated_response.status,
                (await page.content()).encode(),  # NOTE: we can do better
                known_encoding="utf-8",
            )

            return response

    def request(
        self, url: str, raise_on_statuses: Optional[Container[int]] = None
    ) -> Response:
        return self.run_in_default_context(
            self.__request, url, raise_on_statuses=raise_on_statuses
        )



# =============================================================================
# Minet Tiktok Constants
# =============================================================================
#
# General constants used throughout the Tiktok functions.
#
from urllib3 import Timeout

TIKTOK_URL = "https://www.tiktok.com"
TIKTOK_PUBLIC_API_DEFAULT_TIMEOUT = Timeout(connect=10, read=60 * 2)
TIKTOK_DEFAULT_THROTTLE = 3.0
TIKTOK_MIN_TIME_RETRYER = 5.0
TIKTOK_MAX_RANDOM_ADDENDUM = 1.0



# =============================================================================
# Minet YouTube Exceptions
# =============================================================================
#
from minet.exceptions import MinetError


class YouTubeError(MinetError):
    pass


class YouTubeInvalidAPIKeyError(YouTubeError):
    pass


class YouTubeAPILimitReached(YouTubeError):
    pass


class YouTubePotentiallyTransientInvalidAPICallError(YouTubeError):
    def __init__(self, url, status, data):
        super().__init__()
        self.url = url
        self.status = status
        self.data = data

    def __str__(self):
        return super().__str__() + ", Url: %s, Status: %i, Data: %s" % (
            self.url,
            self.status,
            self.data,
        )


class YouTubeInvalidAPICallError(YouTubeError):
    def __init__(self, url, status, data):
        super().__init__()
        self.url = url
        self.status = status
        self.data = data

    def __str__(self):
        return super().__str__() + ", Url: %s, Status: %i, Data: %s" % (
            self.url,
            self.status,
            self.data,
        )


class YouTubeInvalidVideoTargetError(YouTubeError):
    pass


class YouTubeInvalidChannelTargetError(YouTubeError):
    pass


class YouTubeDisabledCommentsError(YouTubeError):
    pass


class YouTubeNotFoundError(YouTubeError):
    pass


class YouTubeExclusiveMemberError(YouTubeError):
    pass


class YouTubeUnknown403Error(YouTubeError):
    pass


class YouTubeAccessNotConfiguredError(YouTubeError):
    def __init__(self, message=None, url=None):
        super().__init__(message)
        self.url = url



# =============================================================================
# Minet Crawler State
# =============================================================================
#
# Simple class representing a miner Crawler's state.
#
from typing import Callable, Optional

from threading import Lock
from contextlib import contextmanager

CrawlerStateListener = Callable[["CrawlerState"], None]


class CrawlerState(object):
    jobs_done: int
    jobs_doing: int
    jobs_queued: int

    listener: Optional[CrawlerStateListener]

    __lock: Lock

    def __init__(
        self,
        jobs_queued: Optional[int] = None,
        listener: Optional[CrawlerStateListener] = None,
    ):
        self.__lock = Lock()

        self.jobs_done = 0
        self.jobs_doing = 0
        self.jobs_queued = jobs_queued if jobs_queued is not None else 0

        self.listener = listener

    @property
    def total(self) -> int:
        return self.jobs_done + self.jobs_doing + self.jobs_queued

    def __notify(self):
        if callable(self.listener):
            self.listener(self)

    def set_listener(self, listener: CrawlerStateListener):
        with self.__lock:
            self.listener = listener

    def inc_queued(self, count=1) -> None:
        with self.__lock:
            self.jobs_queued += count
            self.__notify()

    def dec_queued(self) -> None:
        with self.__lock:
            self.jobs_queued -= 1
            self.__notify()

    def inc_done(self) -> None:
        with self.__lock:
            self.jobs_done += 1
            self.__notify()

    def inc_doing(self) -> None:
        with self.__lock:
            self.jobs_doing += 1
            self.__notify()

    def dec_doing(self) -> None:
        with self.__lock:
            self.jobs_doing -= 1
            self.__notify()

    def inc_working(self) -> None:
        with self.__lock:
            self.jobs_queued -= 1
            self.jobs_doing += 1
            self.__notify()

    def dec_working(self) -> None:
        with self.__lock:
            self.jobs_done += 1
            self.jobs_doing -= 1
            self.__notify()

    @contextmanager
    def task(self):
        try:
            self.inc_working()
            yield
        finally:
            self.dec_working()

    def __repr__(self):
        class_name = self.__class__.__name__

        return (
            "<%(class_name)s queued=%(jobs_queued)i doing=%(jobs_doing)i done=%(jobs_done)i>"
        ) % {
            "class_name": class_name,
            "jobs_done": self.jobs_done,
            "jobs_queued": self.jobs_queued,
            "jobs_doing": self.jobs_doing,
        }



# =============================================================================
# Minet Custom CLI Exceptions
# =============================================================================
#
# Collection of handy custom exceptions.
#
from minet.exceptions import MinetError


class MinetCLIError(MinetError):
    pass


class InvalidArgumentsError(MinetCLIError):
    pass


class MissingColumnError(MinetCLIError):
    def __init__(self, message, column):
        super().__init__(message)
        self.column = column


class NotResumableError(MinetCLIError):
    pass


class FatalError(MinetCLIError):
    def __init__(self, message=None, warning=False):
        self.warning = warning
        super().__init__(message)



from typing import Tuple, List

import json
import random
from os.path import join, dirname

import minet.user_agents.data as data
from minet.exceptions import UserAgentsUpdateError


def cum_sum_generator(probs):
    d = 0
    for p in probs:
        d = d + p
        yield d


def update_user_agents(transient: bool = False) -> None:
    from minet.web import request

    def download_useragentsme_data() -> List[Tuple[float, str]]:
        response = request("https://www.useragents.me")
        json_text = response.soup().force_select_one("#most-common-desktop-useragents-json-csv textarea").get_text()
        data = json.loads(json_text)

        return [
            (row["pct"], row["ua"])
            for row in data
            if not "Trident" in row["ua"]  # Filtering out windows shenanigans etc.
        ]

    try:
        data.USER_AGENTS = download_useragentsme_data()

        probs = [pct for (pct, _) in data.USER_AGENTS]
        data.USER_AGENTS_SHARE_CDF = [p for p in cum_sum_generator(probs)]

        if transient:
            return

        with open(join(dirname(__file__), "data.py"), "w") as f:
            f.write("USER_AGENTS = [\n")
            for pct, ua in data.USER_AGENTS:
                eua = ua.replace('"', r"\"")
                f.write('\t(%f, "%s"),\n' % (pct, eua))
            f.write("]\n\n")

            f.write("USER_AGENTS_SHARE_CDF = [\n")
            for pct in data.USER_AGENTS_SHARE_CDF:
                f.write("\t%f,\n" % pct)
            f.write("]\n\n")

    except Exception as reason:
        raise UserAgentsUpdateError(
            "Unable to update the list of user agents", reason=reason
        )


def get_random_user_agent() -> str:
    return random.choices(
        data.USER_AGENTS, cum_weights=data.USER_AGENTS_SHARE_CDF, k=1
    )[0][1]



# =============================================================================
# Minet Utils
# =============================================================================
#
# Miscellaneous helper function used throughout the library.
#
from typing import Iterator, Any

import os
import re
import sys
import hashlib
import json
import time
import string
import sqlite3
import importlib
from os.path import dirname, abspath, relpath
from random import uniform

from minet.exceptions import (
    GenericModuleNotFoundError,
    TargetInGenericModuleNotFoundError,
)


def fuzzy_int(value):
    try:
        return int(value)
    except ValueError:
        return int(float(value))


def md5(string):
    h = hashlib.md5()
    h.update(string.encode())
    return h.hexdigest()


DOUBLE_QUOTES_RE = re.compile(r'"')


def fix_ensure_ascii_json_string(s):
    try:
        return json.loads('"%s"' % DOUBLE_QUOTES_RE.sub('\\"', s))
    except json.decoder.JSONDecodeError:
        return s


class PseudoFStringFormatter(string.Formatter):
    def get_field(self, field_name, args, kwargs):
        result = eval(field_name, None, kwargs)

        return result, None


def sleep_with_entropy(seconds, max_random_addendum):
    random_addendum = uniform(0, max_random_addendum)
    time.sleep(seconds + random_addendum)


def is_binary_mimetype(m: str) -> bool:
    if m.startswith("text/"):
        return False

    if not m.startswith("application/"):
        return True

    second_part = m.split("/", 1)[-1]

    return not (
        "json" in second_part
        or "html" in second_part
        or "xml" in second_part
        or "yaml" in second_part
        or "yml" in second_part
        or second_part == "x-httpd-php"
    )


NUMBER_RE = re.compile(r"\d+[\.,]?\d*[KM]?")


def clean_human_readable_numbers(text):
    match = NUMBER_RE.search(text)

    if match is None:
        return text

    approx_likes = match.group(0)

    if "K" in approx_likes:
        approx_likes = str(int(float(approx_likes[:-1]) * 10**3))

    elif "M" in approx_likes:
        approx_likes = str(int(float(approx_likes[:-1]) * 10**6))

    approx_likes = approx_likes.replace(",", "")
    approx_likes = approx_likes.replace(".", "")

    return approx_likes


def message_flatmap(*messages, sep=" ", end="\n"):
    return sep.join(
        end.join(m for m in message) if not isinstance(message, str) else message
        for message in messages
    )


def parse_module_and_target(path, default: str = "main"):
    if ":" in path:
        s = path.rsplit(":", 1)
        return s[0], s[1]

    return path, default


def import_target(path: str, default: str = "main"):
    module_path_or_name, function_name = parse_module_and_target(path, default=default)

    # NOTE: we normalize to a path, so we can add dir to sys.path
    if not module_path_or_name.endswith(".py"):
        module_path_or_name = module_path_or_name.replace(".", os.sep) + ".py"

    module_path = abspath(module_path_or_name)
    module_directory = dirname(module_path)

    if module_directory not in sys.path:
        sys.path.append(module_directory)

    # NOTE: we renormalize to a module
    module = relpath(module_path)[:-3].replace(os.sep, ".")

    try:
        m = importlib.import_module(module)
    except ImportError:
        raise GenericModuleNotFoundError(module_path_or_name, path)

    sys.path.remove(module_directory)

    try:
        return getattr(m, function_name)
    except AttributeError:
        raise TargetInGenericModuleNotFoundError(
            module_path_or_name, path, function_name
        )


def iterate_over_sqlite_cursor(cursor: sqlite3.Cursor) -> Iterator[Any]:
    while True:
        rows = cursor.fetchmany(128)

        if not rows:
            return

        for row in rows:
            yield row



# =============================================================================
# Minet Facebook Post Authors CLI Action
# =============================================================================
#
# Logic of the `fb post-authors` action.
#
from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.facebook.utils import with_facebook_fatal_errors
from minet.facebook import FacebookMobileScraper
from minet.facebook.types import MobileFacebookUser
from minet.facebook.exceptions import FacebookInvalidTargetError


@with_facebook_fatal_errors
@with_enricher_and_loading_bar(
    headers=MobileFacebookUser, title="Finding authors", unit="posts"
)
def action(cli_args, enricher, loading_bar):
    scraper = FacebookMobileScraper(cli_args.cookie, throttle=cli_args.throttle)

    for i, row, post_url in enricher.enumerate_cells(
        cli_args.column, with_rows=True, start=1
    ):
        with loading_bar.step():
            try:
                author = scraper.post_author(post_url)
            except FacebookInvalidTargetError:
                loading_bar.print(
                    "Given url (line %i) is probably not a Facebook group post: %s"
                    % (i, post_url)
                )
                continue

            enricher.writerow(row, author.as_csv_row() if author is not None else None)



# =============================================================================
# Minet Mediacloud API Client
# =============================================================================
#
# A unified mediacloud API client that can be used to keep an eye on the
# rate limit and the used token etc.
#
from minet.web import create_pool_manager
from minet.mediacloud.constants import MEDIACLOUD_DEFAULT_TIMEOUT
from minet.mediacloud.utils import make_simple_call
from minet.mediacloud.search import mediacloud_search
from minet.mediacloud.topic import mediacloud_topic_stories
from minet.mediacloud.types import MediacloudFeed, MediacloudMedia


class MediacloudAPIClient(object):
    def __init__(self, token):
        self.token = token
        self.pool_manager = create_pool_manager(timeout=MEDIACLOUD_DEFAULT_TIMEOUT)

    def count(self, query, **kwargs):
        return mediacloud_search(
            self.pool_manager, self.token, query, count=True, **kwargs
        )

    def search(self, query, **kwargs):
        return mediacloud_search(
            self.pool_manager, self.token, query, count=False, **kwargs
        )

    def topic_stories(self, topic_id, **kwargs):
        return mediacloud_topic_stories(
            self.pool_manager, self.token, topic_id, **kwargs
        )

    def media(self, media_id, **kwargs):
        return make_simple_call(
            self.pool_manager,
            self.token,
            "/media/single",
            MediacloudFeed.from_payload,
            arg=media_id,
            single=True,
            **kwargs,
        )

    def feeds(self, media_id, **kwargs):
        return make_simple_call(
            self.pool_manager,
            self.token,
            "/feeds/list",
            MediacloudMedia.from_payload,
            query={"media_id": media_id, "rows": 100},
            **kwargs,
        )



from typing import (
    Union,
    Optional,
    Callable,
    Any,
    cast,
    Dict,
    List,
    get_origin,
    get_args,
    get_type_hints,
)

import inspect

from casanova import RowWrapper
from bs4 import SoupStrainer

from minet.scrape.classes.base import ScraperBase
from minet.scrape.soup import WonderfulSoup
from minet.scrape.straining import strainer_from_css
from minet.scrape.utils import ensure_soup
from minet.scrape.types import AnyScrapableTarget


def infer_fieldnames_from_function_return_type(fn: Callable) -> Optional[List[str]]:
    if not callable(fn):
        raise TypeError

    return_type = get_type_hints(fn)["return"]

    origin = get_origin(return_type)

    if origin is Union:
        args = get_args(return_type)

        # Optionals
        if len(args) == 2:
            if args[1] is type(None):
                return_type = args[0]

    if return_type in (str, int, float, bool, type(None)):
        return ["value"]

    return None


class FunctionScraper(ScraperBase):
    fn: Union[str, Callable[[RowWrapper, WonderfulSoup], Any]]
    fieldnames = None
    plural: bool
    tabular = True
    output_type = None
    strainer: Optional[SoupStrainer]

    def __init__(
        self,
        fn: Union[str, Callable[[RowWrapper, WonderfulSoup], Any]],
        strain: Optional[str] = None,
    ):
        # NOTE: closures cannot be pickled without using third-party library `dill`.
        self.fn = fn
        self.plural = inspect.isgeneratorfunction(fn)

        self.strainer = None

        if strain is not None:
            self.strainer = strainer_from_css(strain)

    def __call__(self, html: AnyScrapableTarget, context: Optional[Dict] = None):
        assert context is not None

        row = context["row"]
        soup = cast(WonderfulSoup, ensure_soup(html, strainer=self.strainer))

        if isinstance(self.fn, str):
            return eval(self.fn, {"row": row, "soup": soup}, None)

        return self.fn(row, soup)



from typing import Literal

from dataclasses import dataclass
from casanova import TabularRecord

Granularity = Literal["daily", "monthly"]
Agent = Literal["user", "all-agents"]
Access = Literal["all-access", "mobile-web", "mobile-app"]


@dataclass
class WikipediaPageViewsItem(TabularRecord):
    project: str
    article: str
    granularity: Granularity
    timestamp: int
    access: Access
    agent: Agent
    views: int



from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.loading_bar import LoadingBar
from minet.youtube.scraper import YouTubeScraper


@with_enricher_and_loading_bar(
    headers=["title", "url"],
    title="Retrieving channel links",
    unit="channels",
    sub_unit="links",
    nested=True,
)
def action(cli_args, enricher, loading_bar: LoadingBar):
    scraper = YouTubeScraper()

    for row, channel_url in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step(channel_url):
            links = scraper.get_channel_links(channel_url)

            if links is None:
                continue

            for title, link in links:
                enricher.writerow(row, [title, link])
                loading_bar.nested_advance()



from typing import List, Optional, Any, Literal

import re
from ural import could_be_html, normalize_url, should_follow_href
from urllib.parse import urljoin
from dataclasses import dataclass
from casanova import TabularRecord

from minet.web import looks_like_html
from minet.crawl.types import CrawlJob, SuccessfulCrawlResult
from minet.extraction import extract
from minet.web import Response
from minet.crawl.spiders import Spider


FocusCrawlContentInvalidity = Literal["irrelevant", "not-html", "binary", "empty"]
FocusCrawlLinkInvalidity = Literal["irrelevant", "not-html"]


@dataclass
class FocusCrawLinkReport(TabularRecord):
    source: str
    target: str
    source_is_relevant: bool
    source_content_invalidity: Optional[FocusCrawlContentInvalidity]
    target_is_relevant: bool
    target_link_invalidity: Optional[FocusCrawlLinkInvalidity]


@dataclass
class FocusCrawLink:
    url: str
    invalidity: Optional[FocusCrawlLinkInvalidity] = None

    @property
    def relevant(self) -> bool:
        return self.invalidity is None


@dataclass
class FocusCrawlInfo:
    CRAWL_RESULT_ADDENDUM_FIELDNAMES = ["relevant", "invalidity", "matches"]

    invalidity: Optional[FocusCrawlContentInvalidity] = None
    matches: Optional[int] = None
    links: Optional[List[FocusCrawLink]] = None

    @property
    def relevant(self) -> bool:
        return self.invalidity is None

    def as_crawl_result_csv_addendum(self):
        return ["yes" if self.relevant else "no", self.invalidity, self.matches]


class FocusSpider(Spider):
    def clean_url(self, origin, url):
        url = urljoin(origin, url)
        return normalize_url(url)

    def url_relevance(self, url):
        if self.regex_url:
            m = bool(self.regex_url.search(url))

            if self.invert_url_match:
                m = not m

            return m

        return True

    def content_relevance(self, content):
        if self.regex_content:
            matches = self.regex_content.findall(content)
            if self.invert_content_match:
                return not bool(matches), None
            else:
                return bool(matches), len(matches) if matches else None
        else:
            return True, None

    def __init__(
        self,
        start_urls=None,
        regex_content=None,
        invert_content_match: bool = False,
        regex_url=None,
        invert_url_match: bool = False,
        irrelevant_continue: bool = False,
        extract: bool = False,
        only_html: bool = True,
        extraction_fields=None,
    ):
        if not regex_content and not regex_url:
            raise TypeError("Neither url nor content filter provided.")

        if extraction_fields and not extract:
            raise TypeError(
                "Custom extraction fields can't be used without the extract option."
            )

        self.urls = start_urls
        self.regex_content = re.compile(regex_content, re.I) if regex_content else None
        self.invert_content_match = invert_content_match
        self.regex_url = re.compile(regex_url, re.I) if regex_url else None
        self.invert_url_match = invert_url_match
        self.extraction = extract
        self.extraction_fields = extraction_fields
        self.irrelevant_continue = irrelevant_continue
        self.target_html = only_html

    def process(self, job: CrawlJob, response: Response):
        has_relevant_content = False
        html = response.body

        if self.target_html and not looks_like_html(html):
            return FocusCrawlInfo(invalidity="not-html"), None
        if not response.is_text:
            return FocusCrawlInfo(invalidity="binary"), None
        if not html:
            return FocusCrawlInfo(invalidity="empty")

        html = response.text()
        soup = response.soup(ignore_xhtml_warning=True)

        content = ""

        if self.extraction:
            extraction = extract(html)
            if extraction:
                content = extraction.blurb(self.extraction_fields)
        else:
            content = soup.get_text()

        a_tags = soup.select("a[href]")
        unique_urls = set()

        for a in a_tags:
            url = a.get("href")

            assert isinstance(url, str)

            if not should_follow_href(url):
                continue

            unique_urls.add(response.urljoin(url))

        has_relevant_content, relevant_size = self.content_relevance(content)

        links = [FocusCrawLink(url=url) for url in unique_urls]
        next_urls = []

        for link in links:
            url = link.url

            if self.target_html and not could_be_html(url):
                link.invalidity = "not-html"
                continue

            if not self.url_relevance(url):
                link.invalidity = "irrelevant"
                continue

            next_urls.append(url)

        if not has_relevant_content and not self.irrelevant_continue:
            next_urls = set()

        info = FocusCrawlInfo(
            invalidity="irrelevant" if not has_relevant_content else None,
            matches=relevant_size,
            links=links,
        )

        return info, next_urls

    def start(self):
        if not self.urls:
            return

        yield from self.urls

    def tabulate(self, result: SuccessfulCrawlResult[Any, FocusCrawlInfo]):
        info = result.data
        source_url = result.response.end_url

        if not info.links:
            return

        for link in info.links:
            yield FocusCrawLinkReport(
                source=source_url,
                target=link.url,
                source_is_relevant=info.relevant,
                source_content_invalidity=info.invalidity,
                target_is_relevant=link.relevant,
                target_link_invalidity=link.invalidity,
            )



from typing import Any, cast

from bs4 import SoupStrainer, BeautifulSoup
from ural import should_follow_href
from urllib.parse import urljoin

from .types import NamedScraper


class ImagesScraper(NamedScraper):
    name = "images"
    fieldnames = ["src"]
    plural = True
    output_type = "list"
    strainer = SoupStrainer(name="img")

    def scrape(self, soup: BeautifulSoup, context=None) -> Any:
        img_elems = soup.select("img[src]")
        base_url = context.get("url") if context is not None else None

        urls = []

        for img in img_elems:
            url = img.get("src")

            if url is None:
                continue

            url = cast(str, url).strip()

            if not url:
                continue

            if not should_follow_href(url):
                continue

            if base_url:
                url = urljoin(base_url, url)

            urls.append(url)

        return urls



# =============================================================================
# Minet Hyphe CLI Action
# =============================================================================
#
# Logic of the `hyphe` action.
#
from argparse import FileType

from minet.cli.argparse import command, SplitterType, InputAction
from minet.cli.crawl import crawl_command

HYPHE_API_URL_ARGUMENT = {"name": "url", "help": "Url of the Hyphe API."}
HYPHE_CORPUS_ARGUMENT = {"name": "corpus", "help": "Id of the corpus."}
HYPHE_PASSWORD_ARGUMENT = {
    "flag": "--password",
    "help": "The corpus's password if required.",
}


def hyphe_corpus_subcommand(*args, arguments=[], **kwargs):
    return command(
        *args,
        arguments=[HYPHE_API_URL_ARGUMENT, HYPHE_CORPUS_ARGUMENT]
        + arguments
        + [HYPHE_PASSWORD_ARGUMENT],
        **kwargs
    )


HYPHE_CRAWL_SUBCOMMAND = crawl_command(
    "crawl",
    "minet.cli.hyphe.crawl",
    title="Minet Hyphe Crawl Command",
    description="""
        Specialized crawl command that can be used to reproduce
        a Hyphe crawl from a corpus exported in CSV.
    """,
    epilog="""
        Examples:

        . Reproducing a crawl:
            $ minet hyphe crawl corpus.csv
    """,
    unique=True,
    accept_input=False,
    default_folder_strategy="fullpath",
    default_throttle=0,
    default_connect_timeout=15,
    default_timeout=60,
    default_retries=3,
    force_spoof_user_agent=True,
    force_stateful_redirects=True,
    arguments=[
        {
            "name": "corpus",
            "action": InputAction,
            "help": "Path to the Hyphe corpus exported to CSV.",
        },
        {
            "flag": "--id-column",
            "default": "ID",
            "help": "Name of the CSV column containing the webentity ids.",
        },
        {
            "flag": "--status-column",
            "default": "STATUS",
            "help": "Name of the CSV column containing the webentity statuses.",
        },
        {
            "flag": "--prefixes-column",
            "default": "PREFIXES AS URL",
            "help": "Name of the CSV column containing the webentity prefixes, separated by --prefix-separator.",
        },
        {
            "flag": "--prefix-separator",
            "default": " ",
            "help": "Separator character for the webentity prefixes.",
        },
        {
            "flag": "--start-pages-column",
            "default": "START PAGES",
            "help": "Name of the CSV column containing the webentity start pages, separated by --start-page-separator.",
        },
        {
            "flag": "--start-page-separator",
            "default": " ",
            "help": "Separator character for the webentity start pages.",
        },
        {
            "flag": "--ignore-internal-links",
            "help": "Whether not to write links internal to a webentity on disk.",
            "action": "store_true",
        },
    ],
)


HYPHE_DECLARE_SUBCOMMAND = hyphe_corpus_subcommand(
    "declare",
    "minet.cli.hyphe.declare",
    title="Minet Hyphe Declare Command",
    description="""
        Command that can be used to declare series of webentities
        in a corpus.

        It is ideal to start or restart a corpus using the same exact
        webentity declarations as another one.
    """,
    epilog="""
        Examples:

        . Declaring webentities from a Hyphe export:
            $ minet hyphe declare http://myhyphe.com/api/ target-corpus export.csv
    """,
    total=True,
    arguments=[
        {
            "name": "webentities",
            "help": "CSV file of webentities (exported from Hyphe).",
            "type": FileType("r", encoding="utf-8"),
        }
    ],
)

HYPHE_DESTROY_SUBCOMMAND = hyphe_corpus_subcommand(
    "destroy",
    "minet.cli.hyphe.destroy",
    title="Minet Hyphe Destroy Command",
    description="""
        Command that can be used to destroy a corpus entirely.
    """,
    epilog="""
        Examples:

        . Destroying a corpus:
            $ minet hyphe destroy http://myhyphe.com/api/ my-corpus
    """,
)

HYPHE_DUMP_SUBCOMMAND = hyphe_corpus_subcommand(
    "dump",
    "minet.cli.hyphe.dump",
    title="Minet Hyphe Dump Command",
    description="""
        Command dumping page-level information from a given
        Hyphe corpus.
    """,
    epilog="""
        Examples:

        . Dumping a corpus into the ./corpus directory:
            $ minet hyphe dump http://myhyphe.com/api/ corpus-name -O corpus
    """,
    arguments=[
        {
            "flags": ["-O", "--output-dir"],
            "help": "Output directory for dumped files. Will default to some name based on corpus name.",
        },
        {
            "flag": "--body",
            "help": "Whether to download pages body.",
            "action": "store_true",
        },
        {
            "flag": "--statuses",
            "help": 'Webentity statuses to dump, separated by comma. Possible statuses being "IN", "OUT", "UNDECIDED" and "DISCOVERED".',
            "type": SplitterType(),
        },
        {
            "flag": "--page-count",
            "help": "Number of pages to download per pagination call. Tweak if corpus has large pages or if the network is unreliable.",
            "type": int,
            "default": 500,
        },
    ],
)

HYPHE_RESET_SUBCOMMAND = hyphe_corpus_subcommand(
    "reset",
    "minet.cli.hyphe.reset",
    title="Minet Hyphe Reset Command",
    description="""
        Command that can be used to reset a corpus entirely.
    """,
    epilog="""
        Examples:

        . Resetting a corpus:
            $ minet hyphe reset http://myhyphe.com/api/ my-corpus
    """,
)

HYPHE_TAG_SUBCOMMAND = hyphe_corpus_subcommand(
    "tag",
    "minet.cli.hyphe.tag",
    title="Minet Hyphe Tag Command",
    description="""
        Command that can be used to tag webentities in batch using
        metadata recorded in a CSV file.
    """,
    epilog="""
        Examples:

        . Tag webentities from two columns of CSV file:
            $ minet hyphe tag http://myhyphe.com/api/ my-corpus webentity_id type,creator metadata.csv
    """,
    total=True,
    arguments=[
        {
            "name": "webentity_id_column",
            "help": "Column of the CSV file containing the webentity ids.",
        },
        {
            "name": "tag_columns",
            "help": "Columns, separated by comma, to use as tags.",
            "type": SplitterType(),
        },
        {
            "name": "data",
            "help": "CSV file of webentities (exported from Hyphe).",
            "type": FileType("r", encoding="utf-8"),
        },
        {
            "flag": "--separator",
            "help": "Separator use to split multiple tag values in the same column.",
            "default": "|",
        },
    ],
)

HYPHE_COMMAND = command(
    "hyphe",
    "minet.cli.hyphe",
    "Minet Hyphe Command",
    description="""
        Commands related to the Hyphe web crawler.
    """,
    subcommands=[
        HYPHE_CRAWL_SUBCOMMAND,
        HYPHE_DECLARE_SUBCOMMAND,
        HYPHE_DESTROY_SUBCOMMAND,
        HYPHE_DUMP_SUBCOMMAND,
        HYPHE_RESET_SUBCOMMAND,
        HYPHE_TAG_SUBCOMMAND,
    ],
)



# =============================================================================
# Minet Twitter API Scraper
# =============================================================================
#
# Twitter public API "scraper".
#
import time
import json
import datetime
from urllib.parse import urlencode, quote
from twitwi import normalize_tweet, normalize_user
from ebbe import with_is_first, getpath, pathgetter
from json import JSONDecodeError
from tenacity import RetryCallState

from minet.web import (
    create_pool_manager,
    request,
    create_request_retryer,
    retrying_method,
)
from minet.cookies import (
    coerce_cookie_for_url_from_browser,
    get_cookie_morsel_value,
    cookie_string_to_dict,
)
from minet.rate_limiting import RateLimiterState, rate_limited_method
from minet.twitter.constants import (
    TWITTER_PUBLIC_API_DEFAULT_TIMEOUT,
    TWITTER_PUBLIC_API_AUTH_HEADER,
)
from minet.twitter.exceptions import (
    TwitterPublicAPIGuestTokenError,
    TwitterPublicAPIBadRequest,
    TwitterPublicAPIRateLimitError,
    TwitterPublicAPIInvalidResponseError,
    TwitterPublicAPIParsingError,
    TwitterPublicAPIQueryTooLongError,
    TwitterPublicAPIOverCapacityError,
    TwitterPublicAPIHiccupError,
    TwitterPublicAPIncompleteTweetIndexError,
    TwitterPublicAPIncompleteUserIndexError,
    TwitterPublicAPIInvalidCookieError,
    TwitterPublicAPIBadAuthError,
    TwitterPublicAPINotWorkingAnymore,
    TwitterPublicAPINotFound,
)

# =============================================================================
# Constants
# =============================================================================
TWITTER_PUBLIC_SEARCH_ENDPOINT = "https://api.twitter.com/2/search/adaptive.json"
TWITTER_GUEST_ACTIVATE_ENDPOINT = "https://api.twitter.com/1.1/guest/activate.json"
MAXIMUM_QUERY_LENGTH = 500
DEFAULT_COUNT = 100  # NOTE: the actual upper limit seems to be 20, but I keep 100 just in case it changes in the future, who knows...
MAX_GUEST_TOKEN_USE_COUNT = 100


# =============================================================================
# Helpers
# =============================================================================
def is_query_too_long(query):
    return len(quote(query)) > MAXIMUM_QUERY_LENGTH


def is_cookie_valid(cookie: str) -> bool:
    dict_cookie = cookie_string_to_dict(cookie)

    return bool(
        dict_cookie.get("guest_id")
        and dict_cookie.get("ct0")
        and dict_cookie.get("auth_token")
    )


def ensure_guest_token(method):
    def wrapped(self: "TwitterGuestAPIScraper", *args, **kwargs):
        # NOTE: we refresh the guest token periodically to avoid
        # losing time wrt failed request after expiration
        if (
            self.guest_token is None
            or self.guest_token_use_count >= MAX_GUEST_TOKEN_USE_COUNT
        ):
            self.acquire_guest_token()
            self.guest_token_use_count = 0

        self.guest_token_use_count += 1
        return method(self, *args, **kwargs)

    return wrapped


def create_cookie_expiration():
    return datetime.datetime.utcfromtimestamp(time.time() + 10800).strftime(
        "%a, %d %b %Y %H:%M:%S GMT"
    )


TWEET_SEARCH_DEFAULT_PARAMS = {
    "include_profile_interstitial_type": "1",
    "include_blocking": "1",
    "include_blocked_by": "1",
    "include_followed_by": "1",
    "include_want_retweets": "1",
    "include_mute_edge": "1",
    "include_can_dm": "1",
    "include_can_media_tag": "1",
    "include_ext_has_nft_avatar": "1",
    "include_ext_is_blue_verified": "1",
    "include_ext_verified_type": "1",
    "skip_status": "1",
    "cards_platform": "Web-12",
    "include_cards": "1",
    "include_ext_alt_text": "true",
    "include_ext_limited_action_results": "false",
    "include_quote_count": "true",
    "include_reply_count": "1",
    "tweet_mode": "extended",
    "include_ext_collab_control": "true",
    "include_ext_views": "true",
    "include_entities": "true",
    "include_user_entities": "true",
    "include_ext_media_color": "true",
    "include_ext_media_availability": "true",
    "include_ext_sensitive_media_warning": "true",
    "include_ext_trusted_friends_metadata": "true",
    "send_error_codes": "true",
    "simple_quoted_tweet": "true",
    "tweet_search_mode": "live",
    "query_source": "spelling_expansion_revert_click",
    "pc": "1",
    "spelling_corrections": "1",
    "include_ext_edit_control": "true",
    "ext": "mediaStats,highlightedLabel,hasNftAvatar,voiceInfo,enrichments,superFollowMetadata,unmentionInfo,editControl,collab_control,vibe",
}

USER_SEARCH_DEFAULT_PARAMS = {
    "include_profile_interstitial_type": "1",
    "include_blocking": "1",
    "include_blocked_by": "1",
    "include_followed_by": "1",
    "include_want_retweets": "1",
    "include_mute_edge": "1",
    "include_can_dm": "1",
    "include_can_media_tag": "1",
    "include_ext_has_nft_avatar": "1",
    "skip_status": "1",
    "cards_platform": "Web-12",
    "include_cards": "1",
    "include_ext_alt_text": "true",
    "include_quote_count": "true",
    "include_reply_count": "1",
    "tweet_mode": "extended",
    "include_entities": "true",
    "include_user_entities": "true",
    "include_ext_media_color": "true",
    "include_ext_media_availability": "true",
    "include_ext_sensitive_media_warning": "true",
    "send_error_codes": "true",
    "simple_quoted_tweet": "true",
    "result_filter": "user",
    "query_source": "typed_query",
    "pc": "1",
    "spelling_corrections": "0",
    "ext": "mediaStats,highlightedLabel,hasNftAvatar,voiceInfo,superFollowMetadata",
}


def forge_search_params(query, target="tweets", count=DEFAULT_COUNT, cursor=None):
    if target == "tweets":
        base = TWEET_SEARCH_DEFAULT_PARAMS
    elif target == "users":
        base = USER_SEARCH_DEFAULT_PARAMS
    else:
        raise TypeError('Invalid target. Expecting "tweets" or "users".')

    params = {**base, "q": query, "count": count}

    if cursor is not None:
        params["cursor"] = cursor

    return urlencode(params, quote_via=quote)


CURSOR_FIRST_POSSIBLE_PATH_GETTER = pathgetter(
    [
        "timeline",
        "instructions",
        0,
        "addEntries",
        "entries",
        -1,
        "content",
        "operation",
        "cursor",
        "value",
    ]
)

CURSOR_SECOND_POSSIBLE_PATH_GETTER = pathgetter(
    [
        "timeline",
        "instructions",
        -1,
        "replaceEntry",
        "entry",
        "content",
        "operation",
        "cursor",
        "value",
    ]
)

CURSOR_THIRD_POSSIBLE_PATH_GETTER = pathgetter(
    [
        "data",
        "search_by_raw_query",
        "search_timeline",
        "timeline",
        "instructions",
        0,
        "entries",
        -1,
        "content",
        "value",
    ]
)

CURSOR_FOURTH_POSSIBLE_PATH_GETTER = pathgetter(
    [
        "data",
        "search_by_raw_query",
        "search_timeline",
        "timeline",
        "instructions",
        -1,
        "entry",
        "content",
        "value",
    ]
)

CURSOR_USER_PATH_GETTER = pathgetter(
    [
        "timeline",
        "instructions",
        -1,
        "addEntries",
        "entries",
        -1,
        "content",
        "operation",
        "cursor",
        "value",
    ]
)

ENTRIES_FOLLOWER_YOU_KNOW_PATH_GETTER = pathgetter(
    ["data", "user", "result", "timeline", "timeline", "instructions", -1, "entries"]
)


def extract_cursor_from_tweets_payload(payload):
    found_cursor = CURSOR_FIRST_POSSIBLE_PATH_GETTER(payload)

    if found_cursor is None:
        found_cursor = CURSOR_SECOND_POSSIBLE_PATH_GETTER(payload)

    if found_cursor is None:
        found_cursor = CURSOR_THIRD_POSSIBLE_PATH_GETTER(payload)

    if found_cursor is None:
        found_cursor = CURSOR_FOURTH_POSSIBLE_PATH_GETTER(payload)

    return found_cursor


def extract_cursor_from_users_payload(payload):
    return CURSOR_USER_PATH_GETTER(payload)


def extract_data_from_followers_you_know_payload(payload):
    entries = ENTRIES_FOLLOWER_YOU_KNOW_PATH_GETTER(payload)

    if entries is None:
        return None, []

    bottom_cursor_node = next(
        (entry for entry in entries if entry["entryId"].startswith("cursor-bottom")),
        None,
    )

    if bottom_cursor_node is None:
        return None, []

    return bottom_cursor_node["content"]["value"], [
        entry["content"]["itemContent"]["user_results"]["result"]
        for entry in entries
        if entry["entryId"].startswith("user-")
    ]


def process_single_tweet(tweet_id, tweet_index, user_index):
    try:
        tweet = tweet_index[tweet_id]
    except KeyError:
        raise TwitterPublicAPIncompleteTweetIndexError(
            tweet_id=tweet_id, tweet_index=tweet_index
        )

    try:
        tweet["user"] = user_index[tweet["user_id_str"]]
    except KeyError:
        raise TwitterPublicAPIncompleteUserIndexError(
            user_id=tweet["user_id_str"], user_index=user_index
        )

    # Quoted?
    quoted_id = tweet.get("quoted_status_id_str")

    # TODO: sometimes tweet is not present
    if quoted_id and quoted_id in tweet_index:
        tweet["quoted_status"] = process_single_tweet(
            quoted_id, tweet_index, user_index
        )

    # Retweeted?
    retweeted_id = tweet.get("retweeted_status_id_str")

    # TODO: sometimes tweet is not present
    if retweeted_id and retweeted_id in tweet_index:
        tweet["retweeted_status"] = process_single_tweet(
            retweeted_id, tweet_index, user_index
        )

    return tweet


def recombobulate_tweet(tweet_root, entry_id=None):
    __typename = tweet_root["__typename"]

    if __typename == "TweetUnavailable":
        return None

    if __typename == "TweetWithVisibilityResults":
        tweet_root = tweet_root["tweet"]

    try:
        tweet_meta = tweet_root["legacy"]
    except KeyError:
        raise TwitterPublicAPIParsingError(entry_id)

    tweet_meta["source"] = tweet_root["source"]

    # NOTE: tombstone not working anymore
    # if tweet_meta is None:
    #     tweet_meta = getpath(
    #         entry, ["content", "item", "content", "tombstone", "tweet"]
    #     )

    # Skipping ads
    if "promotedMetadata" in tweet_meta:
        return None

    tweet_index = None
    user_index = None

    # NOTE: with new API results we need to build this index each time
    tweet_index = {tweet_meta["id_str"]: tweet_meta}
    user_index = {}

    # Tweet user
    user_index[tweet_meta["user_id_str"]] = tweet_root["core"]["user_results"][
        "result"
    ]["legacy"]

    # Quote
    quoted_root = getpath(tweet_root, ["quoted_status_result", "result"])

    if quoted_root is not None:
        if quoted_root["__typename"] == "TweetWithVisibilityResults":
            quoted_root = quoted_root["tweet"]

        quoted_meta = quoted_root["legacy"]
        quoted_meta["source"] = quoted_root["source"]
        tweet_index[quoted_meta["id_str"]] = quoted_meta
        user_index[quoted_meta["user_id_str"]] = quoted_root["core"]["user_results"][
            "result"
        ]["legacy"]

    for user_str_id, user_meta in user_index.items():
        user_meta["id_str"] = user_str_id

    return process_single_tweet(tweet_meta["id_str"], tweet_index, user_index)


def tweets_payload_iter(payload):
    instructions = payload["data"]["search_by_raw_query"]["search_timeline"][
        "timeline"
    ]["instructions"]

    # with open("dump.json", "w") as f:
    #     import json

    #     json.dump(payload, f, ensure_ascii=False, indent=2)

    for instruction in instructions:
        if instruction["type"] == "TimelineAddEntries":
            entries = instruction["entries"]
        elif instruction["type"] == "TimelineReplaceEntry":
            entries = [instruction["entry"]]
        else:
            continue

        for entry in entries:
            entry_id = entry["entryId"]

            # Filtering tweets
            if not entry_id.startswith("sq-I-t-") and not entry_id.startswith("tweet-"):
                continue

            tweet_root = getpath(
                entry, ["content", "itemContent", "tweet_results", "result"]
            )

            # Parsing error?
            # NOTE: new API is inconsistent and returns empty objects...
            if tweet_root is None:
                continue
                # raise TwitterPublicAPIParsingError

            tweet = recombobulate_tweet(tweet_root, entry_id=entry_id)

            if tweet is None:
                continue

            # Additional metadata
            # NOTE: not working anymore
            meta = None

            if tweet is not None:
                # if "forwardPivot" in tweet_meta:
                #     pivot = tweet_meta["forwardPivot"]

                #     meta = {
                #         "intervention_text": getpath(pivot, ["text", "text"]),
                #         "intervention_type": pivot.get("displayType"),
                #         "intervention_url": getpath(pivot, ["landingUrl", "url"]),
                #     }

                yield tweet, meta


# =============================================================================
# Main class
# =============================================================================
class TwitterAPIScraper:
    def __init__(self, cookie: str):
        self.pool_manager = create_pool_manager(
            timeout=TWITTER_PUBLIC_API_DEFAULT_TIMEOUT, spoof_tls_ciphers=True
        )

        # NOTE: 10 calls per minute
        self.rate_limiter_state = RateLimiterState(10, 60)

        # self.reset()

        # NOTE: since 2023-04-21, Twitter search is not available anymore
        # if you are not logged in to the website. This means that the old
        # method of acquiring and rotating a guest token is now moot.
        # An alternative is therefore to rely on the user's cookie to
        # spoof the connection and scrape the search as before.
        #
        # The issue here is that Twitter knows who you are and I strongly
        # expect them to kick accounts scraping in the near future as it
        # would be quite easy to do so.
        #
        # They might also very well limit the number of search result
        # returned to prevent scraping altogether, connected or not.
        self.cookie = coerce_cookie_for_url_from_browser(
            cookie, TWITTER_PUBLIC_SEARCH_ENDPOINT
        )

        if not is_cookie_valid(self.cookie):
            raise TwitterPublicAPIInvalidCookieError

        self.csrf_token = get_cookie_morsel_value(self.cookie, "ct0")

        def epilog_builder(retry_state: RetryCallState):
            exc = retry_state.outcome.exception()

            if hasattr(exc, "format_epilog") and callable(exc.format_epilog):
                return exc.format_epilog()

            return None

        self.retryer = create_request_retryer(
            min=1,
            additional_exceptions=[
                TwitterPublicAPIRateLimitError,
                TwitterPublicAPIInvalidResponseError,
                TwitterPublicAPIOverCapacityError,
                TwitterPublicAPIncompleteTweetIndexError,
                TwitterPublicAPIncompleteUserIndexError,
                TwitterPublicAPIHiccupError,  # TODO: I might want to drop this at some point
            ],
            epilog=epilog_builder,
        )

    # def reset(self):
    #     self.guest_token = None
    #     self.guest_token_use_count = 0
    #     self.cookie = None

    @rate_limited_method()
    def request(self, url, headers=None, method="GET"):
        return request(
            url,
            pool_manager=self.pool_manager,
            spoof_ua=True,
            method=method,
            headers=headers,
        )

    # def acquire_guest_token(self):
    #     headers = {
    #         "Authorization": TWITTER_PUBLIC_API_AUTH_HEADER,
    #         "Accept-Language": "en-US,en;q=0.5",
    #     }

    #     response = self.request(TWITTER_GUEST_ACTIVATE_ENDPOINT, headers, method="POST")

    #     if response.status >= 400:
    #         raise TwitterPublicAPIInvalidResponseError(
    #             status=response.status, response_text=response.text
    #         )

    #     try:
    #         api_token_response = response.json()
    #     except JSONDecodeError:
    #         raise TwitterPublicAPIInvalidResponseError(
    #             status=response.status, response_text=response.text
    #         )

    #     guest_token = api_token_response.get("guest_token")

    #     if guest_token is None:
    #         raise TwitterPublicAPIGuestTokenError

    #     self.guest_token = guest_token
    #     self.cookie = response.headers[
    #         "Set-Cookie"
    #     ] + ", gt=%s; Domain=.twitter.com; Path=/; Secure ; Expires=%s" % (
    #         guest_token,
    #         create_cookie_expiration(),
    #     )

    # @ensure_guest_token
    def api_call(self, url):
        headers = {
            "Authorization": TWITTER_PUBLIC_API_AUTH_HEADER,
            # "X-Guest-Token": self.guest_token,
            "Cookie": self.cookie,
            "Accept-Language": "en-US,en;q=0.5",
            "X-Csrf-Token": self.csrf_token,
            "X-Twitter-Active-User": "yes",
            "X-Twitter-Auth-Type": "OAuth2Session",
            "X-Twitter-Client-Language": "en",
        }

        response = self.request(url, headers=headers)

        if response.status == 429:
            # self.reset()
            raise TwitterPublicAPIRateLimitError

        if response.status in [401, 403]:
            # self.reset()
            raise TwitterPublicAPIBadAuthError(response.status)

        try:
            data = response.json()
        except JSONDecodeError:
            raise TwitterPublicAPIInvalidResponseError(
                status=response.status, response_text=response.text
            )

        if response.status >= 400:
            error = getpath(data, ["errors", 0])

            if error is not None and response.status == 400 and error.get("code") == 47:
                raise TwitterPublicAPIBadRequest

            if error is not None and error.get("code") == 130:
                raise TwitterPublicAPIOverCapacityError

            raise TwitterPublicAPIInvalidResponseError(
                status=response.status, response_text=response.text
            )

        return data

    @retrying_method()
    def request_tweet_search(self, query, locale, cursor=None, refs=None, dump=False):
        query = quote(json.dumps(query, ensure_ascii=False))

        # TODO: refactor lol
        url = "https://twitter.com/i/api/graphql/L1VfBERtzc3VkBBT0YAYHA/SearchTimeline?variables=%7B%22rawQuery%22%3A{}%2C%22count%22%3A20%2C%22querySource%22%3A%22typed_query%22%2C%22product%22%3A%22Latest%22%7D&features=%7B%22rweb_lists_timeline_redesign_enabled%22%3Atrue%2C%22responsive_web_graphql_exclude_directive_enabled%22%3Atrue%2C%22verified_phone_label_enabled%22%3Afalse%2C%22creator_subscriptions_tweet_preview_api_enabled%22%3Atrue%2C%22responsive_web_graphql_timeline_navigation_enabled%22%3Atrue%2C%22responsive_web_graphql_skip_user_profile_image_extensions_enabled%22%3Afalse%2C%22tweetypie_unmention_optimization_enabled%22%3Atrue%2C%22responsive_web_edit_tweet_api_enabled%22%3Atrue%2C%22graphql_is_translatable_rweb_tweet_is_translatable_enabled%22%3Atrue%2C%22view_counts_everywhere_api_enabled%22%3Atrue%2C%22longform_notetweets_consumption_enabled%22%3Atrue%2C%22responsive_web_twitter_article_tweet_consumption_enabled%22%3Afalse%2C%22tweet_awards_web_tipping_enabled%22%3Afalse%2C%22freedom_of_speech_not_reach_fetch_enabled%22%3Atrue%2C%22standardized_nudges_misinfo%22%3Atrue%2C%22tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled%22%3Atrue%2C%22longform_notetweets_rich_text_read_enabled%22%3Atrue%2C%22longform_notetweets_inline_media_enabled%22%3Atrue%2C%22responsive_web_media_download_video_enabled%22%3Afalse%2C%22responsive_web_enhance_cards_enabled%22%3Afalse%7D&fieldToggles=%7B%22withArticleRichContentState%22%3Afalse%7D".format(
            query
        )

        if cursor is not None:
            url = "https://twitter.com/i/api/graphql/L1VfBERtzc3VkBBT0YAYHA/SearchTimeline?variables=%7B%22rawQuery%22%3A{}%2C%22count%22%3A20%2C%22cursor%22%3A%22{}%22%2C%22querySource%22%3A%22typed_query%22%2C%22product%22%3A%22Latest%22%7D&features=%7B%22rweb_lists_timeline_redesign_enabled%22%3Atrue%2C%22responsive_web_graphql_exclude_directive_enabled%22%3Atrue%2C%22verified_phone_label_enabled%22%3Afalse%2C%22creator_subscriptions_tweet_preview_api_enabled%22%3Atrue%2C%22responsive_web_graphql_timeline_navigation_enabled%22%3Atrue%2C%22responsive_web_graphql_skip_user_profile_image_extensions_enabled%22%3Afalse%2C%22tweetypie_unmention_optimization_enabled%22%3Atrue%2C%22responsive_web_edit_tweet_api_enabled%22%3Atrue%2C%22graphql_is_translatable_rweb_tweet_is_translatable_enabled%22%3Atrue%2C%22view_counts_everywhere_api_enabled%22%3Atrue%2C%22longform_notetweets_consumption_enabled%22%3Atrue%2C%22responsive_web_twitter_article_tweet_consumption_enabled%22%3Afalse%2C%22tweet_awards_web_tipping_enabled%22%3Afalse%2C%22freedom_of_speech_not_reach_fetch_enabled%22%3Atrue%2C%22standardized_nudges_misinfo%22%3Atrue%2C%22tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled%22%3Atrue%2C%22longform_notetweets_rich_text_read_enabled%22%3Atrue%2C%22longform_notetweets_inline_media_enabled%22%3Atrue%2C%22responsive_web_media_download_video_enabled%22%3Afalse%2C%22responsive_web_enhance_cards_enabled%22%3Afalse%7D&fieldToggles=%7B%22withArticleRichContentState%22%3Afalse%7D".format(
                query, quote(cursor)
            )

        # NOTE: old method
        # params = forge_search_params(query, cursor=cursor, target="tweets")
        # url = "%s?%s" % (TWITTER_PUBLIC_SEARCH_ENDPOINT, params)

        data = self.api_call(url)

        next_cursor = extract_cursor_from_tweets_payload(data)

        tweets = []

        if dump:
            return data

        for tweet, meta in tweets_payload_iter(data):
            try:
                result = normalize_tweet(
                    tweet,
                    locale=locale,
                    extract_referenced_tweets=refs is not None,
                    collection_source="scraping",
                )
            except Exception:
                raise TwitterPublicAPIParsingError(tweet["id_str"])

            if refs is not None:
                for is_first, extracted_tweet in with_is_first(result):
                    # Casting to int64 to save up memory
                    id_int64 = int(extracted_tweet["id"])

                    if id_int64 in refs:
                        continue

                    if is_first:
                        tweets.append((extracted_tweet, meta))
                    else:
                        tweets.append((extracted_tweet, None))

                    refs.add(id_int64)
            else:
                tweets.append((result, meta))

        # Attempting to fix Twitter's public-facing API recent hiccups (#316):
        # It seems that sometimes the API returns an empty response, containing
        # the same cursor as before, in which case we should retry...
        # NOTE: in fact, the real issue lies elsewhere as the way to hit the API
        # changed slightly and was causing our issues. But we cannot rely
        # on this condition now because it will degenerate to a retry loop
        # when hitting the last available tweets of the query.
        # NOTE: since those API changes, we cannot get 100 results at once anymore
        # as the upper limit seems to be 20 now :'(

        # if not tweets and cursor == next_cursor:
        #     raise TwitterPublicAPIHiccupError

        return next_cursor, tweets

    @retrying_method()
    def request_user_search(self, query, locale, cursor=None, dump=False):
        raise TwitterPublicAPINotWorkingAnymore
        # params = forge_search_params(query, cursor=cursor, target="users")
        # url = "%s?%s" % (TWITTER_PUBLIC_SEARCH_ENDPOINT, params)

        # data = self.api_call(url)

        # next_cursor = extract_cursor_from_users_payload(data)

        # if dump:
        #     return data

        # users = [
        #      (user, locale=locale)
        #     for user in data["globalObjects"]["users"].values()
        # ]

        # return next_cursor, users

    @retrying_method()
    def request_followers_you_know(self, user_id, cursor=None):
        url = "https://twitter.com/i/api/graphql/sT_K0qB2bqpWfL3-cq4zaQ/FollowersYouKnow?variables=%7B%22userId%22%3A%22{}%22%2C%22count%22%3A20%2C%22includePromotedContent%22%3Afalse%7D&features=%7B%22responsive_web_graphql_exclude_directive_enabled%22%3Atrue%2C%22verified_phone_label_enabled%22%3Afalse%2C%22creator_subscriptions_tweet_preview_api_enabled%22%3Atrue%2C%22responsive_web_graphql_timeline_navigation_enabled%22%3Atrue%2C%22responsive_web_graphql_skip_user_profile_image_extensions_enabled%22%3Afalse%2C%22c9s_tweet_anatomy_moderator_badge_enabled%22%3Atrue%2C%22tweetypie_unmention_optimization_enabled%22%3Atrue%2C%22responsive_web_edit_tweet_api_enabled%22%3Atrue%2C%22graphql_is_translatable_rweb_tweet_is_translatable_enabled%22%3Atrue%2C%22view_counts_everywhere_api_enabled%22%3Atrue%2C%22longform_notetweets_consumption_enabled%22%3Atrue%2C%22responsive_web_twitter_article_tweet_consumption_enabled%22%3Afalse%2C%22tweet_awards_web_tipping_enabled%22%3Afalse%2C%22freedom_of_speech_not_reach_fetch_enabled%22%3Atrue%2C%22standardized_nudges_misinfo%22%3Atrue%2C%22tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled%22%3Atrue%2C%22rweb_video_timestamps_enabled%22%3Atrue%2C%22longform_notetweets_rich_text_read_enabled%22%3Atrue%2C%22longform_notetweets_inline_media_enabled%22%3Atrue%2C%22responsive_web_media_download_video_enabled%22%3Afalse%2C%22responsive_web_enhance_cards_enabled%22%3Afalse%7D".format(
            quote(user_id)
        )

        if cursor is not None:
            url = "https://twitter.com/i/api/graphql/sT_K0qB2bqpWfL3-cq4zaQ/FollowersYouKnow?variables=%7B%22userId%22%3A%22{}%22%2C%22count%22%3A20%2C%22cursor%22%3A%22{}%22%2C%22includePromotedContent%22%3Afalse%7D&features=%7B%22responsive_web_graphql_exclude_directive_enabled%22%3Atrue%2C%22verified_phone_label_enabled%22%3Afalse%2C%22creator_subscriptions_tweet_preview_api_enabled%22%3Atrue%2C%22responsive_web_graphql_timeline_navigation_enabled%22%3Atrue%2C%22responsive_web_graphql_skip_user_profile_image_extensions_enabled%22%3Afalse%2C%22c9s_tweet_anatomy_moderator_badge_enabled%22%3Atrue%2C%22tweetypie_unmention_optimization_enabled%22%3Atrue%2C%22responsive_web_edit_tweet_api_enabled%22%3Atrue%2C%22graphql_is_translatable_rweb_tweet_is_translatable_enabled%22%3Atrue%2C%22view_counts_everywhere_api_enabled%22%3Atrue%2C%22longform_notetweets_consumption_enabled%22%3Atrue%2C%22responsive_web_twitter_article_tweet_consumption_enabled%22%3Afalse%2C%22tweet_awards_web_tipping_enabled%22%3Afalse%2C%22freedom_of_speech_not_reach_fetch_enabled%22%3Atrue%2C%22standardized_nudges_misinfo%22%3Atrue%2C%22tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled%22%3Atrue%2C%22rweb_video_timestamps_enabled%22%3Atrue%2C%22longform_notetweets_rich_text_read_enabled%22%3Atrue%2C%22longform_notetweets_inline_media_enabled%22%3Atrue%2C%22responsive_web_media_download_video_enabled%22%3Afalse%2C%22responsive_web_enhance_cards_enabled%22%3Afalse%7D".format(
                quote(user_id), quote(cursor)
            )

        data = self.api_call(url)

        next_cursor, user_entries = extract_data_from_followers_you_know_payload(data)

        users = []

        for user_entry in user_entries:
            user_data = user_entry["legacy"]
            user_data["id_str"] = user_entry["rest_id"]
            user_data["protected"] = user_data.get("protected", False)
            users.append(user_data)

        return next_cursor, users

    def search_tweets(
        self,
        query,
        locale=None,
        limit=None,
        include_referenced_tweets=False,
        with_meta=False,
    ):
        if is_query_too_long(query):
            raise TwitterPublicAPIQueryTooLongError

        cursor = None
        i = 0

        refs = set() if include_referenced_tweets else None

        while True:
            new_cursor, tweets = self.request_tweet_search(
                query, locale, cursor, refs=refs
            )

            if not tweets:
                return

            for tweet, meta in tweets:
                if with_meta:
                    yield tweet, meta

                else:
                    yield tweet

                i += 1

                if limit is not None and i >= limit:
                    return

            # NOTE: we cannot stop when no tweets are found anymore because
            # of Twitter's public facing API's strange hiccups.
            # Comparing cursors seems to be the only good option now...
            if new_cursor is None or new_cursor == cursor:
                return

            cursor = new_cursor

    def search_users(self, query, locale=None, limit=None):
        if is_query_too_long(query):
            raise TwitterPublicAPIQueryTooLongError

        cursor = None
        i = 0

        while True:
            new_cursor, users = self.request_user_search(query, locale, cursor)

            if not users:
                return

            for user in users:
                yield user

                i += 1

                if limit is not None and i >= limit:
                    return

            if new_cursor is None or new_cursor == cursor:
                return

            cursor = new_cursor

    def followers_you_know(self, user_id, locale=None):
        cursor = None

        while True:
            next_cursor, users = self.request_followers_you_know(user_id, cursor=cursor)

            if not users:
                return

            for user in users:
                yield normalize_user(user, locale=locale)

            if next_cursor is None or next_cursor == cursor:
                return

            cursor = next_cursor


class TwitterGuestAPIScraper:
    def __init__(self):
        self.pool_manager = create_pool_manager(
            timeout=TWITTER_PUBLIC_API_DEFAULT_TIMEOUT, spoof_tls_ciphers=True
        )

        # NOTE: 1 call per 2 second
        self.rate_limiter_state = RateLimiterState(1, 2)

        self.reset()

        def epilog_builder(retry_state: RetryCallState):
            exc = retry_state.outcome.exception()

            if hasattr(exc, "format_epilog") and callable(exc.format_epilog):
                return exc.format_epilog()

            return None

        self.retryer = create_request_retryer(
            min=1,
            additional_exceptions=[
                TwitterPublicAPIRateLimitError,
                TwitterPublicAPIInvalidResponseError,
                TwitterPublicAPIOverCapacityError,
                TwitterPublicAPIncompleteTweetIndexError,
                TwitterPublicAPIncompleteUserIndexError,
                TwitterPublicAPIHiccupError,  # TODO: I might want to drop this at some point
            ],
            epilog=epilog_builder,
        )

    def reset(self):
        self.guest_token = None
        self.guest_token_use_count = 0
        self.cookie = None

    @rate_limited_method()
    def request(self, url, headers=None, method="GET"):
        return request(
            url,
            pool_manager=self.pool_manager,
            spoof_ua=True,
            method=method,
            headers=headers,
        )

    def acquire_guest_token(self):
        headers = {
            "Authorization": TWITTER_PUBLIC_API_AUTH_HEADER,
            "Accept-Language": "en-US,en;q=0.5",
        }

        response = self.request(TWITTER_GUEST_ACTIVATE_ENDPOINT, headers, method="POST")

        if response.status >= 400:
            raise TwitterPublicAPIInvalidResponseError(
                status=response.status, response_text=response.text
            )

        try:
            api_token_response = response.json()
        except JSONDecodeError:
            raise TwitterPublicAPIInvalidResponseError(
                status=response.status, response_text=response.text
            )

        guest_token = api_token_response.get("guest_token")

        if guest_token is None:
            raise TwitterPublicAPIGuestTokenError

        self.guest_token = guest_token
        self.cookie = response.headers[
            "Set-Cookie"
        ] + ", gt=%s; Domain=.twitter.com; Path=/; Secure ; Expires=%s" % (
            guest_token,
            create_cookie_expiration(),
        )

    @ensure_guest_token
    def api_call(self, url):
        headers = {
            "Authorization": TWITTER_PUBLIC_API_AUTH_HEADER,
            "X-Guest-Token": self.guest_token,
            "Cookie": self.cookie,
            "Accept-Language": "en-US,en;q=0.5",
            "X-Twitter-Active-User": "yes",
            "X-Twitter-Client-Language": "en",
        }

        response = self.request(url, headers=headers)

        if response.status == 429:
            self.reset()
            raise TwitterPublicAPIRateLimitError

        if response.status in [401, 403]:
            self.reset()
            raise TwitterPublicAPIBadAuthError(response.status)

        try:
            data = response.json()
        except JSONDecodeError:
            raise TwitterPublicAPIInvalidResponseError(
                status=response.status, response_text=response.text
            )

        if response.status >= 400:
            error = getpath(data, ["errors", 0])

            if error is not None and response.status == 400 and error.get("code") == 47:
                raise TwitterPublicAPIBadRequest

            if error is not None and error.get("code") == 130:
                raise TwitterPublicAPIOverCapacityError

            raise TwitterPublicAPIInvalidResponseError(
                status=response.status, response_text=response.text
            )

        return data

    @retrying_method()
    def tweet(self, tweet_id, locale=None):
        url = "https://api.twitter.com/graphql/8be30APApp7lds22UzVS7Q/TweetResultByRestId?variables=%7B%22tweetId%22%3A%22{}%22%2C%22withCommunity%22%3Afalse%2C%22includePromotedContent%22%3Afalse%2C%22withVoice%22%3Afalse%7D&features=%7B%22creator_subscriptions_tweet_preview_api_enabled%22%3Atrue%2C%22c9s_tweet_anatomy_moderator_badge_enabled%22%3Atrue%2C%22tweetypie_unmention_optimization_enabled%22%3Atrue%2C%22responsive_web_edit_tweet_api_enabled%22%3Atrue%2C%22graphql_is_translatable_rweb_tweet_is_translatable_enabled%22%3Atrue%2C%22view_counts_everywhere_api_enabled%22%3Atrue%2C%22longform_notetweets_consumption_enabled%22%3Atrue%2C%22responsive_web_twitter_article_tweet_consumption_enabled%22%3Afalse%2C%22tweet_awards_web_tipping_enabled%22%3Afalse%2C%22freedom_of_speech_not_reach_fetch_enabled%22%3Atrue%2C%22standardized_nudges_misinfo%22%3Atrue%2C%22tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled%22%3Atrue%2C%22rweb_video_timestamps_enabled%22%3Atrue%2C%22longform_notetweets_rich_text_read_enabled%22%3Atrue%2C%22longform_notetweets_inline_media_enabled%22%3Atrue%2C%22responsive_web_graphql_exclude_directive_enabled%22%3Atrue%2C%22verified_phone_label_enabled%22%3Afalse%2C%22responsive_web_media_download_video_enabled%22%3Afalse%2C%22responsive_web_graphql_skip_user_profile_image_extensions_enabled%22%3Afalse%2C%22responsive_web_graphql_timeline_navigation_enabled%22%3Atrue%2C%22responsive_web_enhance_cards_enabled%22%3Afalse%7D".format(
            quote(tweet_id)
        )

        data = self.api_call(url)

        tweet_root = getpath(data, ("data", "tweetResult", "result"))

        if tweet_root is None:
            raise TwitterPublicAPINotFound(tweet_id)

        tweet = recombobulate_tweet(tweet_root)
        tweet = normalize_tweet(
            tweet,
            locale=locale,
            collection_source="scraping",
        )

        return tweet



from typing import Dict, Type
from .types import NamedScraper

from .canonical import CanonicalScraper
from .europresse import EuropresseScraper
from .images import ImagesScraper
from .metas import MetasScraper
from .rss import RssScraper
from .title import TitleScraper
from .urls import UrlsScraper

NAMED_SCRAPERS: Dict[str, Type[NamedScraper]] = {
    s.name: s
    for s in [
        TitleScraper,
        CanonicalScraper,
        UrlsScraper,
        ImagesScraper,
        MetasScraper,
        RssScraper,
        EuropresseScraper,
    ]
}



from typing import Any

from bs4 import SoupStrainer, BeautifulSoup

from .types import NamedScraper


class MetasScraper(NamedScraper):
    name = "metas"
    fieldnames = [
        "name",
        "property",
        "http-equiv",
        "itemprop",
        "content",
        "charset",
    ]
    plural = True
    output_type = "collection"
    strainer = SoupStrainer(name="meta")

    def scrape(self, soup: BeautifulSoup, context=None) -> Any:
        meta_elems = soup.find_all(name="meta")

        metas = []

        for meta_elem in meta_elems:
            metas.append({name: meta_elem.get(name) for name in self.fieldnames})

        return metas
